
#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_sdm.h"
#include "app_error.h"
#include "ble.h"
#include "ble_err.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_bas.h"
#include "ble_dis.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "app_timer.h"
#include "bsp_btn_ble.h"
#include "bsp_config.h"
#include "peer_manager.h"
#include "peer_manager_handler.h"
#include "fds.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_lesc.h"
#include "nrf_ble_qwr.h"
#include "ble_conn_state.h"
#include "nrf_pwr_mgmt.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "ble_peripheral.h"


#define BATTERY_LEVEL_MEAS_INTERVAL         APP_TIMER_TICKS(1000)
#define NOTIF_INTERVAL                      APP_TIMER_TICKS(330)
#define KEEP_ALIVE_INTERVAL                 APP_TIMER_TICKS(500)

#define DEAD_BEEF                           0xDEADBEEF                              /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

#define OSTIMER_WAIT_FOR_QUEUE              2                                       /**< Number of ticks to wait for the timer queue to be ready */


APP_TIMER_DEF(m_battery_timer);
APP_TIMER_DEF(m_notification_timer);
APP_TIMER_DEF(m_keep_alive_timer);


void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


static void battery_level_meas_timeout_handler(void * p_context)
{
	UNUSED_PARAMETER(p_context);
	ret_code_t err_code;

	static uint8_t  battery_level = 100;

	err_code = ble_bas_battery_level_update(get_bas_p(), battery_level, BLE_CONN_HANDLE_ALL);
	if ((err_code != NRF_SUCCESS) &&
		(err_code != NRF_ERROR_INVALID_STATE) &&
		(err_code != NRF_ERROR_RESOURCES) &&
		(err_code != NRF_ERROR_BUSY) &&
		(err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
	   )
	{
		APP_ERROR_HANDLER(err_code);
	}
	battery_level--;
}

static void control_keep_alive_timeout_handler(void * p_context)
{
    UNUSED_PARAMETER(p_context);
    bsp_board_led_invert(BSP_LED_INDICATE_USER_LED2);
}

static void timers_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_battery_timer,
    							APP_TIMER_MODE_REPEATED,
								battery_level_meas_timeout_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_notification_timer,
    							APP_TIMER_MODE_REPEATED,
								notif_timeout_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_keep_alive_timer,
    							APP_TIMER_MODE_REPEATED,
								control_keep_alive_timeout_handler);
    APP_ERROR_CHECK(err_code);
}


static void application_timers_start(void)
{
    ret_code_t err_code;
//    err_code = app_timer_start(m_battery_timer, BATTERY_LEVEL_MEAS_INTERVAL, NULL);
//    APP_ERROR_CHECK(err_code);

    err_code = app_timer_start(m_notification_timer, NOTIF_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_start(m_keep_alive_timer, KEEP_ALIVE_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);
}


static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}


static void buttons_leds_init(bool * p_erase_bonds)
{
    ret_code_t err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LEDS | BSP_INIT_BUTTONS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

    *p_erase_bonds = (startup_event == BSP_EVENT_CLEAR_BONDING_DATA);
}

static void idle_state_handle(void)
{
    ret_code_t err_code;

    err_code = nrf_ble_lesc_request_handler();
    APP_ERROR_CHECK(err_code);

    if (NRF_LOG_PROCESS() == false)
    {
        nrf_pwr_mgmt_run();
    }
}

static void power_management_init(void)
{
    ret_code_t err_code;
    err_code = nrf_pwr_mgmt_init();
    APP_ERROR_CHECK(err_code);
}


int main(void)
{
    bool erase_bonds;

    log_init();
    timers_init();
    buttons_leds_init(&erase_bonds);
    power_management_init();
    config_peripheral_ble();

    application_timers_start();
    advertising_start(erase_bonds);

    NRF_LOG_INFO("Control Main Started.");

    for (;;)
    {
        idle_state_handle();
    }
}


