#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "nordic_common.h"
#include "nrf_sdm.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "nrf_pwr_mgmt.h"
#include "app_util.h"
#include "app_error.h"
#include "sdk_config.h"
#include "app_util.h"
#include "app_timer.h"
#include "bsp_btn_ble.h"
#include "fds.h"
#include "nrf_fstorage.h"
#include "nrf_ble_lesc.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_ble_scan.h"
#include "bsp_config.h"

#include "ble_central.h"

#define NOTIF_INTERVAL                      APP_TIMER_TICKS(1000)
#define KEEP_ALIVE_INTERVAL                 APP_TIMER_TICKS(500)


APP_TIMER_DEF(m_notification_timer);
APP_TIMER_DEF(m_keep_alive_timer);

void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

static void service_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}

static void bas_c_init(void)
{
    ble_bas_c_init_t bas_c_init_obj;

    bas_c_init_obj.evt_handler   = bas_c_evt_handler;
    bas_c_init_obj.error_handler = service_error_handler;
    bas_c_init_obj.p_gatt_queue  = get_gatt_queue_p();

    ret_code_t err_code = ble_bas_c_init(get_ble_bas_p(), &bas_c_init_obj);
    APP_ERROR_CHECK(err_code);
}


static void buttons_leds_init(bool * p_erase_bonds)
{
    ret_code_t  err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LEDS | BSP_INIT_BUTTONS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

    *p_erase_bonds = (startup_event == BSP_EVENT_CLEAR_BONDING_DATA);
}


static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

static void lure_keep_alive_timeout_handler(void *p_context)
{
    UNUSED_PARAMETER(p_context);
    bsp_board_led_invert(BSP_LED_INDICATE_USER_LED2);
}

static void timer_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_notification_timer,
    							APP_TIMER_MODE_REPEATED,
								write_timeout_handler);
    APP_ERROR_CHECK(err_code);



    err_code = app_timer_create(&m_keep_alive_timer,
    							APP_TIMER_MODE_REPEATED,
								lure_keep_alive_timeout_handler);
    APP_ERROR_CHECK(err_code);
}

static void application_timers_start(void)
{
	ret_code_t err_code = app_timer_start(m_notification_timer, NOTIF_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_start(m_keep_alive_timer, KEEP_ALIVE_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);
}

static void power_management_init(void)
{
    ret_code_t err_code = nrf_pwr_mgmt_init();
    APP_ERROR_CHECK(err_code);
}

static void idle_state_handle(void)
{
    ret_code_t err_code;

    err_code = nrf_ble_lesc_request_handler();
    APP_ERROR_CHECK(err_code);

    NRF_LOG_FLUSH();
    nrf_pwr_mgmt_run();
}

int main(void)
{
    bool erase_bonds;

    log_init();
    timer_init();
    power_management_init();
    buttons_leds_init(&erase_bonds);
    ble_central_config();
    bas_c_init();

    application_timers_start();
    scanning_start(&erase_bonds);

    NRF_LOG_INFO("Lure App Started.");
    for (;;)
    {
        idle_state_handle();
    }
}


